/**
 * Checks that the passed array equals this one at each index.
 *
 * @name equals
 * @methodOf Array#
 * @param {Array} a The array to check
 * @return {Boolean} true if equal, false if not
 */
var arrayEquals = function (a, b) {
  if (a.length !== b.length) {
    return false;
  }
  for (var i = 0, len = a.length; i < len; i++) {
    if (!objValuesEqual(a[i], b[i])) {
      return false;
    }
  }
  return true;
};

Nativelib.merge(Array.prototype, {
  /**
   * Checks if all items in this array pass the provided test.
   *
   * @name all
   * @methodOf Array#
   * @param {Function} fn The filtering function.
   * @param {Object} ctx The function's execution context.
   * @return {Boolean} If each item passes.
   */
  all: function (fn, ctx) {
    for (var i = 0, len = this.length; i < len; i++) {
      if (!fn.call(ctx || this, this[i], i, this)) {
        return false;
      }
    }
    return true;
  },

  /**
   * Creates a copy of this array.
   *
   * @name copy
   * @methodOf Array#
   * @return {Array} The copy.
   */
  copy: function () {
    return this.slice(0);
  },

  equals: function (a) {
    return arrayEquals(this, a);
  },

  /**
   * Creates an array containing only the elements in this one for
   * which the filtering function returns true.
   *
   * @name filter
   * @methodOf Array#
   * @param {Function} fn The filtering function.
   * @param {Object} ctx The function's execution context.
   * @return {Array} The filtered array.
   */
  filter: function (fn, ctx) {
    var filtered = [];
    for (var i = 0, len = this.length; i < len; i++) {
      if (fn.call(ctx || this, this[i], i, this)) {
        filtered.push(this[i]);
      }
    }
    return filtered;
  },

  /**
   * Creates a flattened array, merging down all sub-arrays.
   *
   * @name flatten
   * @methodOf Array#
   * @return {Array} The flattened array.
   */
  flatten: function () {
    var flattened = [];
    for (var i = 0, len = this.length; i < len; i++) {
      if (testArray(this[i])) {
        flattened = flattened.concat(this[i].flatten());
      } else {
        flattened.push(this[i]);
      }
    }
    return flattened;
  },

  /**
   * The same as filter, but only returning the first match.
   *
   * @name filterOne
   * @methodOf Array#
   * @param {Function} fn The filtering function.
   * @param {Object} ctx The function's execution context.
   * @return {Object} The first filtered item, if found.
   */
  filterOne: function (fn, ctx) {
    for (var i = 0, len = this.length; i < len; i++) {
      if (fn.call(ctx || this, this[i], i, this)) {
        return this[i];
      }
    }
  },

  /**
   * Gets the array's first item.
   *
   * @name first
   * @methodOf Array#
   * @return {Object} The first item.
   */
  first: function () {
    return this[0];
  },

  /**
   * Gets a propertie's value for each object in the array.
   *
   * @name eachProperty
   * @methodOf Array#
   * @param {String} p The property value to retrieve.
   * @param {Function} fn true if the property is a function and should be
   * called.
   * @param {Object} ctx The function's execution context.
   * @return {Array} Each property's value.
   */
  eachProperty: function (p, fn, ctx) {
    var vals = [];
    for (var i = 0, len = this.length; i < len; i++) {
      vals.push(fn ? this[i][p].apply(ctx || this[i][p]) : this[i][p]);
    }
    return vals;
  },

  /**
   * Runs the provided function for each item in this array.
   *
   * @name forEach
   * @methodOf Array#
   * @param {Function} fn The function to run.
   * @param {Object} ctx The function's execution context.
   */
  forEach: function (fn, ctx) {
    for (var i = 0, len = this.length; i < len; i++) {
      fn.call(ctx || this, this[i], i, this);
    }
  },

  /**
   * Checks if the array includes an item identical to the one passed.
   *
   * @name includes
   * @methodOf Array#
   * @param {Object} item The item to check.
   * @return {Boolean} true if the item was found, false if not.
   */
  includes: function (item) {
    for (var i = 0, len = this.length; i < len; i++) {
      if (this[i] === item) {
        return true;
      }
    }
    return false;
  },

  /**
   * Gets all but the last item of the array.
   *
   * @name init
   * @methodOf Array#
   * @return {Array} The array's init.
   */
  init: function () {
    return this.slice(0, this.length - 1);
  },

  /**
   * Calls the property, if present, on each item in this array,
   * returning a list of results. If the named property is not present
   * on a given item, undefined will pushed on the result list.
   *
   * @name invoke
   * @methodOf Array#
   * @param {String} name The property name to call.
   * @param {Array} args An argument list that will be provided with
   * each call.
   * @param {Object} ctx The function's execution context.
   * @return {Array} The return value of each call.
   */
  invoke: function (name, args, ctx) {
    var results = [];
    for (var i = 0, len = this.length; i < len; i++) {
      results.push(name in this[i] ?
        this[i][name].apply(ctx || this[i], args || []) : undefined);
    }
    return results;
  },

  /**
   * Gets the last item of the array.
   *
   * @name last
   * @methodOf Array#
   * @return {Object} The array's last item.
   */
  last: function () {
    return this[this.length - 1];
  },

  /**
   * Creates an array containing the values of a function applied to
   * each item in the array.
   *
   * @example
   * [1, 2, 3].map(function (n, i, a) {
   *   return n * i;
   * }; // [0, 2, 6]
   *
   * @name map
   * @methodOf Array#
   * @param {Function} fn The map function
   * @param {Object} ctx The function's execution context, called with (value,
   * index, array)
   * @return {Array} The mapped array
   */
  map: function (fn, ctx) {
    var mapped = [];
    for (var i = 0, len = this.length; i < len; i++) {
      mapped.push(fn.call(ctx || this, this[i], i, this));
    }
    return mapped;
  },

  /**
   * Splits this array in two based on the return value of the
   * partitioning function. The first array contains all positives and
   * the second all negatives.
   *
   * @name partition
   * @methodOf Array#
   * @param {Function} fn The partition function.
   * @param {Object} ctx The function's execution context.
   * @return {Array} The array of paritions.
   */
  partition: function (fn, ctx) {
    var parts = [[], []];
    for (var i = 0, len = this.length; i < len; i++) {
      parts[fn.call(ctx || this, this[i], i, this) ? 0 : 1].push(this[i]);
    }
    return parts;
  },

  /**
   * Processes the items of an array from left to right, applying the
   * reducing function to each item.
   *
   * @name reduce
   * @methodOf Array#
   * @param {Function} fn The function that is called with each iteration. This
   * function is provided with the following arguments:
   * - The result of the previous iteration
   * - The current item
   * - The current item's index
   * - The array
   * @param {Object} init A value to be used as the previous value in the first
   * iteration.
   * @param {Object} ctx The context in which to apply the function.
   * @return {Object} The reduced value.
   */
  reduce: function (fn, init, ctx) {
    var i = 0, len = this.length, prev = init || this[i++];
    while (i < len) {
      prev = fn.apply(ctx || this, [prev, this[i], i++, this]);
    }
    return prev;
  },

  /**
   * Removes all items matching the passed object.
   *
   * @name remove
   * @methodOf Array#
   * @param {Object} o The object to remove.
   * @return {Boolean} true if the object was removed, false if not.
   */
  remove: function (o) {
    var removed = false;
    for (var i = 0, len = this.length; i < len; i++) {
      if (this[i] === o) {
        removed = true;
        this.splice(i--, 1);
      }
    }
    return removed;
  },

  /**
   * Checks if any item in this array passes the provided test.
   *
   * @name some
   * @methodOf Array#
   * @param {Function} fn The filtering function.
   * @param {Object} ctx The function's execution context.
   * @return {Boolean} If any item passes.
   */
  some: function (fn, ctx) {
    for (var i = 0, len = this.length; i < len; i++) {
      if (fn.call(ctx || this, this[i], i, this)) {
        return true;
      }
    }
    return false;
  },

  /**
   * Returns an array containing all items not identical to those in the
   * passed array.
   *
   * @name subtract
   * @methodOf Array#
   * @param {Array} a The array of of objects to filter.
   * @param {Boolean} v true to check value rather than identity.
   * @return {Array} The array with none of the passed objects.
   */
  subtract: function (a, v) {
    var subtracted = [];

    if (v) {
      for (var i = 0, len = this.length; i < len; i++) {
        for (var j = 0, len2 = a.length; j < len2; j++) {
          if (this[i] != a[j]) {
            subtracted.push(this[i]);
          }
        }
      }
    } else {
      for (var i = 0, len = this.length; i < len; i++) {
        for (var j = 0, len2 = a.length; j < len2; j++) {
          if (this[i] !== a[j]) {
            subtracted.push(this[i]);
          }
        }
      }
    }

    return subtracted;
  },

  /**
   * Returns an array containing only items with unique identity.
   *
   * @name unique
   * @methodOf Array#
   * @return {Array} The array with no duplicate objects.
   */
  unique: function () {
    var uniq = [];
    for (var i = 0, len = this.length; i < len; i++) {
      if (!uniq.includes(this[i])) {
        uniq.push(this[i]);
      }
    }
    return uniq;
  },

  /**
   * Creates an array containing arrays of each item at each index.
   *
   * @example
   * [1, 2, 3].zip([4, 5, 6]); // [[1, 4], [2, 5], [3, 6]]
   *
   * @name zip
   * @methodOf Array#
   * @param {Array} arr The array to zip.
   * @return {Array} The zipped array.
   */
  zip: function (arr) {
    var zipped = [];
    var len = this.length > arr.length ? this.length : arr.length;
    for (var i = 0; i < len; i++) {
      zipped.push([this[i], arr[i]]);
    }
    return zipped;
  }
});

objMerge(Array, {
  /**
   * Creates an array from an array-like object.
   *
   * @name create
   * @methodOf Array
   * @param {Object} o The source object
   * @return {Array} The new array
   */
  create: function (o) {
    var arr;
    if (testArray(o)) {
      arr = o;
    } else if (o.toArray) {
      arr = o.toArray();
    } else {
      arr = [];
      for (var i = 0, len = o.length || 0; i < len; i++) {
        arr.push(o[i]);
      }
    }
    return arr;
  },

  /**
   * Checks if the object is an array.
   *
   * @name test
   * @methodOf Array
   * @param {Object} o The object to check.
   * @return {Boolean} true if it is an array, false if not.
   */
  test: testArray
});

Nativelib.alias(Array.prototype, {
  /**
   * Alias of eachProperty.
   *
   * @name pluck
   * @methodOf Array#
   */
  eachProperty: 'pluck',

  /**
   * Alias of filter.
   *
   * @name select
   * @methodOf Array#
   */
  filter: 'select',

  /**
   * Alias of filterOne.
   *
   * @name detect
   * @methodOf Array#
   */
  filterOne: 'detect',

  /**
   * Alias of forEach.
   *
   * @name each
   * @methodOf Array#
   */
  forEach: 'each'
});
