Obj.merge(Function.prototype, {
  /**
   * Binds an execution context and an optional argument list to this
   * function.
   *
   * @name bind
   * @methodOf Function#
   * @param {Object} ctx The bound execution context.
   * @param {Array} args The bound arguments array.
   * @return {Function} The bound function.
   */
  bind: function (ctx, args) {
    var fn = this;
    return function () {
      return fn.apply(ctx, args || []);
    };
  },

  /**
   * Bind a number of arguments to this function, not specifying
   * execution context.
   *
   * @name curry
   * @methodOf Function#
   * @param {Arguments} list An argument list to bind.
   * @return {Function} The bound function.
   */
  curry: function () {
    var fn = this, args = Array.create(arguments);
    return function () {
      return fn.apply(this, args.concat(Array.create(arguments)));
    };
  },

  /**
   * Calls the function after the specified timeout.
   *
   * @name delay
   * @methodOf Function#
   * @param {Number} ms The number of milliseconds to delay.
   * @param {Object} ctx The function's execution context.
   * @param {Array} args An arguments array to pass to the function.
   * @return {Number} The timeout's ID.
   */
  delay: function (ms, ctx, args) {
    var fn = this;
    return setTimeout(function () {
      fn.apply(ctx || this, args || []);
    }, ms);
  },

  /**
   * Calls this function n times, providing the current iteration
   * (1-based) and n and the each time.
   *
   * @name repeat
   * @methodOf Function#
   * @param {Number} n The number of iterations.
   * @param {Object} ctx (optional) The exection context.
   */
  repeat: function (n, ctx) {
    var fn = this;
    return function (x, n) {
      return n.range(1).reduce(function (p) {
        return fn.call(ctx || this, p);
      }, x);
    }
  }
});

Obj.merge(Function, {
  /**
   * Returns an empty (typically placeholder) function.
   *
   * @name empty
   * @methodOf Function
   * @return {Function} The empty function.
   */
  empty: function () {
    return function () {};
  },

  /**
   * Checks if the provided object is a Function.
   *
   * @name test
   * @methodOf Function
   * @param {Object} o The object to check.
   * @return {Boolean} true if it is a Function, false if not.
   */
  test: function (o) {
    return Object.prototype.toString.apply(o) === '[object Function]';
  }
});

Obj.alias(Function.prototype, {
  /**
   * Alias of curry.
   *
   * @name partial
   * @methodOf Function#
   */
  curry: 'partial'
});
