/**
 * @namespace Functions for dealing with Objects.
 */
Obj = {};

/**
 * Adds all of the passed properties if they are not already
 * defined.
 *
 * @example
 * var o = { a: 1 };
 * Object.merge(o, {
 *   a: 2,
 *   b: 2
 * });
 * o; // { a: 1, b: 2 }
 *
 * @name merge
 * @methodOf Obj
 * @param {Object} o The object to modify (in-place).
 * @param {Object} props The properties to write.
 * @param {Boolean} clobber true to overwrite defined properties
 * @return {Object} The modified object.
 */
var objMerge = function (o, props, clobber) {
  for (var p in props) {
    if (props.hasOwnProperty(p) && (clobber ||
        typeof o[p] === 'undefined')) {
      o[p] = props[p];
    }
  }
  return o;
};

/**
 * Maps each property to its value.
 *
 * @example
 * var o = { a: 1 };
 * Object.alias(o, {
 *   a: 'b'
 * }; // { a: 1, b: 1 }
 * Object.alias(o, {
 *   b: ['c', 'd']
 * }; // { a: 1, b: 1, c: 1, d: 1 }
 *
 * @name alias
 * @methodOf Obj
 * @param {Object} o The object to modify (in-place).
 * @param {Object} map The properties to map.
 * @param {Boolean} clobber true to overwrite defined properties.
 * @return {Object} The modified object.
 */
var objAlias = function (o, map, clobber) {
  for (var p in map) {
    if (map.hasOwnProperty(p)) {
      if (testArray(map[p])) {
        for (var i = 0, len = map[p].length; i < len; i++) {
          if (typeof o[map[p][i]] === 'undefined' || clobber) {
            o[map[p][i]] = o[p];
          }
        }
      } else {
        if (typeof o[map[p]] === 'undefined' || clobber) {
          o[map[p]] = o[p];
        }
      }
    }
  }

  return o;
};

/**
 * Copies each property of the provided object into a new object,
 * creating a shallow copy.
 *
 * @name copy
 * @methodOf Obj
 * @param {Object} o The object to copy.
 * @return {Object} The copied object.
 */
var objCopy = function (o) {
  return objMerge({}, o);
};

/**
 * Creates a new object from an array of property-value pairs.
 *
 * @example
 * Object.create([['a', 1], ['b', 2]]); // { a: 1, b: 2 }
 *
 * @name create
 * @methodOf Obj
 * @param {Array} pairs The pair array.
 * @return {Object} The created object.
 */
var objCreate = function (pairs) {
  var o = {};
  for (var i = 0; i < pairs.length; i++) {
    o[pairs[i][0]] = pairs[i][1];
  }
  return o;
};

/**
 * Compares the value of each property in each object.
 *
 * @name equals
 * @methodOf Obj
 * @param {Object} o One object to compare.
 * @param {Object} o2 The object to compare it with.
 * @return {Boolean} true if the properties are equal, false if not.
 */
var objEquals = function (o, o2) {
  var p;
  for (p in o2) {
    if (o2.hasOwnProperty(p) && !objValuesEqual(o[p], o2[p])) {
      return false;
    }
  }
  for (p in o) {
    if (o.hasOwnProperty(p) && !objValuesEqual(o[p], o2[p])) {
      return false;
    }
  }
  return true;
};

/**
 * Creates a new object containing only the properties that return
 * true when passed to the provided function.
 *
 * @name filter
 * @methodOf Obj
 * @param {Object} o The object to filter.
 * @param {Function} fn The filtering function.
 * @param {Object} ctx The function's execution context.
 * @return {Object} The filtered object.
 */
var objFilter = function (o, fn, ctx) {
  var filtered = {};
  for (var p in o) {
    if (o.hasOwnProperty(p) && fn.call(ctx || o, o[p], p, o)) {
      filtered[p] = o[p];
    }
  }
  return filtered;
};

/**
 * The same as filter, but only returning the first match.
 *
 * @name filterOne
 * @methodOf Obj
 * @param {Object} o The object to filter.
 * @param {Function} fn The filtering function.
 * @param {Object} ctx The function's execution context.
 * @return The first filtered item, if found.
 */
var objFilterOne = function (o, fn, ctx) {
  var filtered = {};
  for (var p in o) {
    if (o.hasOwnProperty(p) && fn.call(ctx || o, o[p], p, o)) {
      filtered[p] = o[p];
      return filtered;
    }
  }
};

/**
 * Returns an object matching the specified path, if it exists.
 *
 * @name fromPath
 * @methodOf Obj
 * @param {String} path The period '.' separated object path.
 * @param {Object} root The root path (default is window).
 * @return {Object} The object, if found.
 */
var objFromPath = function (path, root) {
  return path.split('.').reduce(function (a, b) {
    return a ? a[b] : undefined;
  }, root || window);
};

/**
 * Calls the property, if present, on each property in the passed
 * object, returning an object with the results at each property. If
 * the named property is not present on a given item, the result
 * property will be undefined.
 *
 * @name invoke
 * @methodOf Obj
 * @param {Object} o The object
 * @param {String} name The property name to call.
 * @param {Array} args An argument list that will be provided with
 * each call.
 * @param {Object} ctx The function's execution context.
 * @return {Object} The return value of each call.
 */
var objInvoke = function (o, name, args, ctx) {
  var invoked = {};
  for (var p in o) {
    if (o.hasOwnProperty(p)) {
      invoked[p] = (name in o[p] ?
        o[p][name].apply(ctx || o[p], args || []) : undefined);
    }
  }
  return invoked;
};

/**
 * Creates an array containing the property names of an object.
 *
 * @name keys
 * @methodOf Obj
 * @param {Object} o The object
 * @return {Array} The object's keys
 */
var objKeys = function (o) {
  var keys = [];
  for (var p in o) {
    if (o.hasOwnProperty(p)) {
      keys.push(p);
    }
  }
  return keys;
};

/**
 * Runs the passed function for each property in the object.
 *
 * @name forEach
 * @methodOf Obj
 * @param {Object} o The object to iterate.
 * @param {Function} fn The function to call.
 * @param {Object} ctx The function's execution context.
 */
var objForEach = function (o, fn, ctx) {
  for (var p in o) {
    if (o.hasOwnProperty(p)) {
      fn.call(ctx || o, o[p], p, o);
    }
  }
};

/**
 * Creates an array containing the values of a function applied to
 * each property in the object.
 *
 * @example
 * Object.map({ Rice: 'Jerry', Clark: 'Gary' }, function (v, p, o) {
 *   return p + ', ' + v;
 * }; // { Rice: 'Rice, Jerry', Clark: 'Clark, Gary' }
 *
 * @name map
 * @methodOf Obj
 * @param {Object} o The object to iterate.
 * @param {Function} fn The map function.
 * @param {Object} ctx The function's execution context, called with (value,
 * property, object).
 * @return {Object} The mapped object.
 */
var objMap = function (o, fn, ctx) {
  var mapped = {};
  for (var p in o) {
    if (o.hasOwnProperty(p)) {
      mapped[p] = fn.call(ctx || o, o[p], o, o);
    }
  }
  return mapped;
};

/**
 * Creates a nested object to the depth specified, not overwriting
 * existing properties.
 *
 * @name namespace
 * @methodOf Obj
 * @param {String} path The period '.' separated object path.
 * @return {Object} The resulting object.
 */
var objNamespace = function (path) {
  return path.split('.').reduce(function (o, part) {
    return o[part] || (o[part] = {});
  }, window);
};

/**
 * Checks if the passed item is an object.
 *
 * @name test
 * @methodOf Obj
 * @param {Object} o The item to check.
 * @return {Boolean} true if it is an object, false if not.
 */
var objTest = function (o) {
  return Object.prototype.toString.call(o) === '[object Object]';
};

/**
 * Returns each property's value.
 *
 * @name values
 * @methodOf Obj
 * @param {Object} o The object to use.
 * @return {Array} An array of values.
 */
var objValues = function (o) {
  var vals = [];
  for (var p in o) {
    if (o.hasOwnProperty(p)) {
      vals.push(o[p]);
    }
  }
  return vals;
};

/**
 * Checks if the two passed objects are equal, performing deep
 * checks for Objects and Arrays.
 *
 * @name valuesEqual
 * @methodOf Obj
 * @param {Object} a The item to check.
 * @param {Object} b The item to check it against.
 * @return {Boolean} true if the values are equal, false if not.
 */
var objValuesEqual = function (a, b) {
  if (testArray(a)) {
    return testArray(b) && arrayEquals(a, b);
  } else if (objTest(a)) {
    return objTest(b) && objEquals(a, b);
  } else {
    return a === b;
  }
};

objMerge(Obj, {
  alias: objAlias,
  copy: objCopy,
  create: objCreate,
  equals: objEquals,
  detect: objFilterOne,
  /**
   * Alias of forEach.
   *
   * @name each
   * @methodOf Obj
   */
  each: objForEach,
  filter: objFilter,
  /**
   * Alias of filterOne.
   *
   * @name detect
   * @methodOf Obj
   */
  filterOne: objFilterOne,
  fromPath: objFromPath,
  forEach: objForEach,
  invoke: objInvoke,
  keys: objKeys,
  map: objMap,
  merge: objMerge,
  namespace: objNamespace,
  /**
   * Alias of namespace
   *
   * @name ns
   * @methodOf Obj
   */
  ns: objNamespace,
  /**
   * Alias of filter.
   *
   * @name select
   * @methodOf Obj
   */
  select: objFilter,
  test: objTest,
  values: objValues,
  valuesEqual: objValuesEqual
});
