Nativelib.merge(String.prototype, {
  /**
   * Capitalizes the string's first letter.
   *
   * @name capitalize
   * @methodOf String#
   * @return {String} The capitalized string.
   */
  capitalize: function () {
    return this.substr(0, 1).toUpperCase() + this.slice(1);
  },

  /**
   * Checks if this string ends with the passed string.
   *
   * @name endsWith
   * @methodOf String#
   * @param {String} str The string to check.
   * @return {Boolean} true if it matches, false if not.
   */
  endsWith: function (str) {
    return this.slice(this.length - str.length) === str;
  },

  /**
   * Checks if this string has non-whitespace characters.
   *
   * @name empty
   * @methodOf String#
   * @return {Boolean} true if empty, false, if not.
   */
  empty: function () {
    return this.strip().length < 1;
  },

  /**
   * Creates a formatted string.
   *
   * @example
   * '{0}-{1}-{2}'.format('a', 'b', 'c'); // 'a-b-c'
   * '{1}-{1}-{3}'.format('a', 'b', 'c', 'd'); // 'b-b-d'
   *
   * @name format
   * @methodOf String#
   * @param {Arguments} list An argument list of ordered replacements.
   * @return {String} The formatted string.
   */
  format: function () {
    var replacements = arguments;
    return this.replace(/\{(\d+)\}/g, function (match, group) {
      return replacements[Number(group)];
    });
  },

  /**
   * Concatenates this string n times.
   *
   * @name mult
   * @methodOf String#
   * @param {Number} n The number of times to concatenate the string.
   * @return {String} The multipled string.
   */
  mult: function (n) {
    var str = '';
    for (var i = 0; i < n; i++) {
      str += this;
    }
    return str;
  },

  slugify: function (keepCase) {
    var slug = this.strip().replace(/[^-A-Za-z0-9\s]/g, '').replace(/\s+/g, '-');
    if (!keepCase) {
      slug = slug.toLowerCase();
    }
    return slug;
  },

  /**
   * sprintf implementation: works for basic replacement but
   * precision, width, and flags are not implemented fully.
   *
   * @name sprintf
   * @methodOf String#
   * @param {Arguments} An argument list of ordered replacements.
   * @return {String} The formatted string.
   */
  sprintf: function () {
    var args = Array.prototype.slice.call(arguments);

    return this.replace(String.sprintfRegex, function (match,
      mainGroup, flag, width, precisionMatch, precision,
      specifier, chars) {
        // Special cases that use no argument
        if (specifier === '%') {
          return '%';
        }
        if (specifier === 'n') {
          return chars;
        }
        if (width === '*') {
          width = args.shift();
        }
        if (precisionMatch && precision === '') {
          precision = 0;
        } else if (precision === '*') {
          precision = args.shift();
        }

        var replace = args.shift();

        switch (specifier) {
          case 'c':
          case 's':
            replace = String(replace);
            break;

          case 'd':
          case 'i':
          case 'u':
            replace = Number(replace).floor();
            break;

          case 'e':
            replace = Number(replace).toExponential();
            break;

          case 'E':
            replace = Number(replace).toExponential().toUpperCase();
            break;

          case 'f':
            replace = Number(replace);
            break;

          case 'g':
          if (Number(replace).toExponential().length <=
             String(Number(replace)).length) {
              replace = Number(replace).toExponential();
            } else {
              replace = Number(replace);
            }
            break;

          case 'G':
            if (Number(replace).toExponential().length <=
               String(Number(replace)).length) {
              replace = Number(replace).toExponential().toUpperCase();
            } else {
              replace = Number(replace);
            }
            break;

          case 'o':
            replace = Number(replace).toString(8);
            break;

          case 'x':
            replace = Number(replace).toString(16);
            break;

          case 'X':
            replace = Number(replace).toString(16).toUpperCase();
            break;

          default:
            break;
        }

        return replace;
      }
    );
  },

  /**
   * Checks if this string starts with the passed string.
   *
   * @name startsWith
   * @methodOf String#
   * @param {String} str The string to check.
   * @return {Boolean} true if it matches, false if not.
   */
  startsWith: function (str) {
    return this.slice(0, str.length) === str;
  },

  /**
   * Strips all whitespace characters from each side of the string.
   *
   * @name strip
   * @methodOf String#
   * @return {String} The stripped string.
   */
  strip: function () {
    return this.replace(/^\s+/g, '').replace(/\s+$/, '');
  },

  /**
   * Splits the string into words.
   *
   * @name words
   * @methodOf String#
   * @return {Array} An array of words.
   */
  words: function () {
    return this.split(/\s+/);
  }
});

Nativelib.merge(String, {
  sprintfRegex: new RegExp(
    '%' + // % Prefix
    '(' +
      '([-+ #0])?'       + // Flags
      '(\\d+|\\*)?'      + // Width
      '(\\.(\\d+|\\*))?'   + // Precision
      '([cdieEfgGosuxXpn%])' + // Specifier
    ')',
  'g'),

  /**
   * Checks if the provided object is a string.
   *
   * @name test
   * @methodOf String
   * @param {Object} o The object to check.
   * @return {Boolean} true if it is a string, false if not.
   */
  test: function (o) {
    return typeof o === 'string';
  }
});

Nativelib.alias(String.prototype, {
  /**
   * Alias of mult.
   *
   * @name x
   * @methodOf String#
   */
  mult: 'x'
});
